=== modified file 'loggerhead/load_test.py'
--- loggerhead/load_test.py	2011-02-09 21:17:32 +0000
+++ loggerhead/load_test.py	2011-02-09 23:36:53 +0000
@@ -63,6 +63,7 @@
 this.
 """
 
+import threading
 import time
 import Queue
 
@@ -71,6 +72,12 @@
 except ImportError:
     import json as simplejson
 
+from bzrlib import (
+    errors,
+    transport,
+    urlutils,
+    )
+
 
 class RequestDescription(object):
     """Describes info about a request."""
@@ -80,7 +87,7 @@
         self.relpath = descrip_dict['relpath']
 
 
-class RequestThread(object):
+class RequestWorker(object):
     """Process requests in a worker thread."""
 
     _timer = time.time
@@ -93,11 +100,11 @@
         self.blocking_time = blocking_time
 
     def step_next(self):
-        item = self.queue.get(True, self.blocking_time)
+        url = self.queue.get(True, self.blocking_time)
         self.start_time = self._timer()
-        self.process(item)
+        success = self.process(url)
         self.end_time = self._timer()
-        self.update_stats(item)
+        self.update_stats(url, success)
         self.queue.task_done()
 
     def run(self, stop_event):
@@ -107,26 +114,32 @@
             except Queue.Empty:
                 pass
 
-    def join(self):
-        """Wait until all requests are finished."""
-        self.queue.join()
-
-    def process(self, item):
-        pass
-
-    def update_stats(self, item):
-        self.stats.append((item, self.end_time - self.start_time))
+    def process(self, url):
+        base, path = urlutils.split(url)
+        t = transport.get_transport(base)
+        try:
+            content = t.get_bytes(path)
+        except (errors.TransportError, errors.NoSuchFile):
+            return False
+        return True
+
+    def update_stats(self, url, success):
+        self.stats.append((url, success, self.end_time - self.start_time))
 
 
 class ActionScript(object):
     """This tracks the actions that we want to perform."""
 
-    _thread_class = RequestThread
+    _worker_class = RequestWorker
+    _default_base_url = 'http://localhost:8080'
+    _default_blocking_timeout = 60.0
 
     def __init__(self):
-        self.base_url = 'http://localhost:8080'
-        self.blocking_timeout = 60.0
-        self.requests = []
+        self.base_url = self._default_base_url
+        self.blocking_timeout = self._default_blocking_timeout
+        self._requests = []
+        self._threads = {}
+        self.stop_event = threading.Event()
 
     @classmethod
     def parse(cls, content):
@@ -147,6 +160,48 @@
         if blocking_timeout is not None:
             script.blocking_timeout = blocking_timeout
         for request_dict in request_list:
-            request = RequestDescription(request_dict)
-            script.requests.append(request)
+            script.add_request(request_dict)
         return script
+
+    def add_request(self, request_dict):
+        request = RequestDescription(request_dict)
+        self._requests.append(request)
+
+    def _get_worker(self, thread_id):
+        if thread_id in self._threads:
+            return self._threads[thread_id][0]
+        handler = self._worker_class(thread_id,
+                                     blocking_time=self.blocking_timeout)
+
+        t = threading.Thread(target=handler.run, args=(self.stop_event,),
+                             name='Thread-%s' % (thread_id,))
+        self._threads[thread_id] = (handler, t)
+        t.start()
+        return handler
+
+    def finish_queues(self):
+        """Wait for all queues of all children to finish."""
+        for h, t in self._threads.itervalues():
+            h.queue.join()
+
+    def stop_and_join(self):
+        """Stop all running workers, and return.
+
+        This will stop even if workers still have work items.
+        """
+        self.stop_event.set()
+        for h, t in self._threads.itervalues():
+            # And join the controlling thread
+            t.join(self.blocking_timeout)
+
+    def _full_url(self, relpath):
+        return self.base_url + relpath
+
+    def run(self):
+        self.stop_event.clear()
+        for request in self._requests:
+            full_url = self._full_url(request.relpath)
+            worker = self._get_worker(request.thread)
+            worker.queue.put(full_url)
+        self.finish_queues()
+        self.stop_and_join()

=== modified file 'loggerhead/tests/test_load_test.py'
--- loggerhead/tests/test_load_test.py	2011-02-09 21:17:32 +0000
+++ loggerhead/tests/test_load_test.py	2011-02-09 23:36:53 +0000
@@ -12,11 +12,13 @@
 
 """Tests for the load testing code."""
 
+import socket
 import time
 import threading
 import Queue
 
 from bzrlib import tests
+from bzrlib.tests import http_server
 
 from bzrlib.plugins.loggerhead.loggerhead import load_test
 
@@ -39,56 +41,6 @@
         self.assertEqual('/bar', rd.relpath)
 
 
-class TestActionScript(tests.TestCase):
-
-    def test_parse_requires_parameters_and_requests(self):
-        self.assertRaises(ValueError,
-            load_test.ActionScript.parse, '')
-        self.assertRaises(ValueError,
-            load_test.ActionScript.parse, '{}')
-        self.assertRaises(ValueError,
-            load_test.ActionScript.parse, '{"parameters": {}}')
-        self.assertRaises(ValueError,
-            load_test.ActionScript.parse, '{"requests": []}')
-        self.assertRaises(ValueError,
-            load_test.ActionScript.parse,
-                '{"parameters": {}, "requests": [], "garbage": "section"}')
-        script = load_test.ActionScript.parse(
-            empty_script)
-        self.assertIsNot(None, script)
-
-    def test_parse_default_base_url(self):
-        script = load_test.ActionScript.parse(empty_script)
-        self.assertEqual('http://localhost:8080', script.base_url)
-
-    def test_parse_find_base_url(self):
-        script = load_test.ActionScript.parse(
-            '{"parameters": {"base_url": "http://example.com"},'
-            ' "requests": []}')
-        self.assertEqual('http://example.com', script.base_url)
-
-    def test_parse_default_blocking_timeout(self):
-        script = load_test.ActionScript.parse(empty_script)
-        self.assertEqual(60.0, script.blocking_timeout)
-
-    def test_parse_find_blocking_timeout(self):
-        script = load_test.ActionScript.parse(
-            '{"parameters": {"blocking_timeout": 10.0},'
-            ' "requests": []}'
-        )
-        self.assertEqual(10.0, script.blocking_timeout)
-
-    def test_parse_finds_requests(self):
-        script = load_test.ActionScript.parse(
-            '{"parameters": {}, "requests": ['
-            ' {"relpath": "/foo"},'
-            ' {"relpath": "/bar"}'
-            ' ]}')
-        self.assertEqual(2, len(script.requests))
-        self.assertEqual("/foo", script.requests[0].relpath)
-        self.assertEqual("/bar", script.requests[1].relpath)
-
-
 _cur_time = time.time()
 def one_sec_timer():
     """Every time this timer is called, it increments by 1 second."""
@@ -97,41 +49,43 @@
     return _cur_time
 
 
-class NoopRequestThread(load_test.RequestThread):
+class NoopRequestWorker(load_test.RequestWorker):
 
     # Every call to _timer will increment by one
     _timer = staticmethod(one_sec_timer)
 
     # Ensure that process never does anything
-    def process(self, item):
-        pass
-
-
-class TestRequestThread(tests.TestCase):
+    def process(self, url):
+        return True
+
+
+class TestRequestWorkerInfrastructure(tests.TestCase):
+    """Tests various infrastructure bits, without doing actual requests."""
 
     def test_step_next_tracks_time(self):
-        rt = NoopRequestThread('id')
+        rt = NoopRequestWorker('id')
         rt.queue.put('item')
         rt.step_next()
         self.assertTrue(rt.queue.empty())
-        self.assertEqual([('item', 1.0)], rt.stats)
+        self.assertEqual([('item', True, 1.0)], rt.stats)
 
     def test_step_multiple_items(self):
-        rt = NoopRequestThread('id')
+        rt = NoopRequestWorker('id')
         rt.queue.put('item')
         rt.step_next()
         rt.queue.put('next-item')
         rt.step_next()
         self.assertTrue(rt.queue.empty())
-        self.assertEqual([('item', 1.0), ('next-item', 1.0)], rt.stats)
+        self.assertEqual([('item', True, 1.0), ('next-item', True, 1.0)],
+                         rt.stats)
 
     def test_step_next_will_timeout(self):
         # We don't want step_next to block forever
-        rt = NoopRequestThread('id', blocking_time=0.001)
+        rt = NoopRequestWorker('id', blocking_time=0.001)
         self.assertRaises(Queue.Empty, rt.step_next)
 
     def test_run_stops_for_stop_event(self):
-        rt = NoopRequestThread('id', blocking_time=0.001, _queue_size=2)
+        rt = NoopRequestWorker('id', blocking_time=0.001, _queue_size=2)
         rt.queue.put('item1')
         rt.queue.put('item2')
         event = threading.Event()
@@ -149,9 +103,167 @@
         rt.queue.put('item4')
         rt.queue.put('item5')
         t.join()
-        self.assertEqual([('item1', 1.0), ('item2', 1.0), ('item3', 1.0)],
+        self.assertEqual([('item1', True, 1.0), ('item2', True, 1.0),
+                          ('item3', True, 1.0)],
                          rt.stats[:3])
         # The last event might be item4 or might be item3, the important thing
         # is that even though there are still queued events, we won't
         # process anything past the first
         self.assertNotEqual('item5', rt.stats[-1][0])
+
+
+class TestRequestWorker(tests.TestCaseWithTransport):
+
+    def setUp(self):
+        super(TestRequestWorker, self).setUp()
+        self.transport_readonly_server = http_server.HttpServer
+
+    def test_request_items(self):
+        rt = load_test.RequestWorker('id', blocking_time=0.01, _queue_size=2)
+        self.build_tree(['file1', 'file2'])
+        readonly_url1 = self.get_readonly_url('file1')
+        self.assertStartsWith(readonly_url1, 'http://')
+        readonly_url2 = self.get_readonly_url('file2')
+        rt.queue.put(readonly_url1)
+        rt.queue.put(readonly_url2)
+        rt.step_next()
+        rt.step_next()
+        self.assertEqual(readonly_url1, rt.stats[0][0])
+        self.assertEqual(readonly_url2, rt.stats[1][0])
+
+    def test_request_nonexistant_items(self):
+        rt = load_test.RequestWorker('id', blocking_time=0.01, _queue_size=2)
+        readonly_url1 = self.get_readonly_url('no-file1')
+        rt.queue.put(readonly_url1)
+        rt.step_next()
+        self.assertEqual(readonly_url1, rt.stats[0][0])
+        self.assertEqual(False, rt.stats[0][1])
+
+    def test_no_server(self):
+        s = socket.socket()
+        # Bind to a port, but don't listen on it
+        s.bind(('localhost', 0))
+        url = 'http://%s:%s/path' % s.getsockname()
+        rt = load_test.RequestWorker('id', blocking_time=0.01, _queue_size=2)
+        rt.queue.put(url)
+        rt.step_next()
+        self.assertEqual((url, False), rt.stats[0][:2])
+
+
+class NoActionScript(load_test.ActionScript):
+
+    _thread_class = NoopRequestWorker
+    _default_blocking_timeout = 0.01
+
+
+class TestActionScriptInfrastructure(tests.TestCase):
+
+    def test_parse_requires_parameters_and_requests(self):
+        self.assertRaises(ValueError,
+            load_test.ActionScript.parse, '')
+        self.assertRaises(ValueError,
+            load_test.ActionScript.parse, '{}')
+        self.assertRaises(ValueError,
+            load_test.ActionScript.parse, '{"parameters": {}}')
+        self.assertRaises(ValueError,
+            load_test.ActionScript.parse, '{"requests": []}')
+        self.assertRaises(ValueError,
+            load_test.ActionScript.parse,
+                '{"parameters": {}, "requests": [], "garbage": "section"}')
+        script = load_test.ActionScript.parse(
+            empty_script)
+        self.assertIsNot(None, script)
+
+    def test_parse_default_base_url(self):
+        script = load_test.ActionScript.parse(empty_script)
+        self.assertEqual('http://localhost:8080', script.base_url)
+
+    def test_parse_find_base_url(self):
+        script = load_test.ActionScript.parse(
+            '{"parameters": {"base_url": "http://example.com"},'
+            ' "requests": []}')
+        self.assertEqual('http://example.com', script.base_url)
+
+    def test_parse_default_blocking_timeout(self):
+        script = load_test.ActionScript.parse(empty_script)
+        self.assertEqual(60.0, script.blocking_timeout)
+
+    def test_parse_find_blocking_timeout(self):
+        script = load_test.ActionScript.parse(
+            '{"parameters": {"blocking_timeout": 10.0},'
+            ' "requests": []}'
+        )
+        self.assertEqual(10.0, script.blocking_timeout)
+
+    def test_parse_finds_requests(self):
+        script = load_test.ActionScript.parse(
+            '{"parameters": {}, "requests": ['
+            ' {"relpath": "/foo"},'
+            ' {"relpath": "/bar"}'
+            ' ]}')
+        self.assertEqual(2, len(script._requests))
+        self.assertEqual("/foo", script._requests[0].relpath)
+        self.assertEqual("/bar", script._requests[1].relpath)
+
+    def test__get_worker(self):
+        script = NoActionScript()
+        # If an id is found, then we should create it
+        self.assertEqual({}, script._threads)
+        worker = script._get_worker('id')
+        self.assertTrue('id' in script._threads)
+        # We should have set the blocking timeout
+        self.assertEqual(script.blocking_timeout,
+                         worker.blocking_time)
+
+        # Another request will return the identical object
+        self.assertIs(worker, script._get_worker('id'))
+
+        # And the stop event will stop the thread
+        script.stop_and_join()
+
+    def test__full_url(self):
+        script = NoActionScript()
+        self.assertEqual('http://localhost:8080/path',
+                         script._full_url('/path'))
+        self.assertEqual('http://localhost:8080/path/to/foo',
+                         script._full_url('/path/to/foo'))
+        script.base_url = 'http://example.com'
+        self.assertEqual('http://example.com/path/to/foo',
+                         script._full_url('/path/to/foo'))
+        script.base_url = 'http://example.com/base'
+        self.assertEqual('http://example.com/base/path/to/foo',
+                         script._full_url('/path/to/foo'))
+        script.base_url = 'http://example.com'
+        self.assertEqual('http://example.com:8080/path',
+                         script._full_url(':8080/path'))
+
+    def test_single_threaded(self):
+        script = NoActionScript.parse("""{
+            "parameters": {"base_url": ""},
+            "requests": [
+                {"thread": "1", "relpath": "first"},
+                {"thread": "1", "relpath": "second"},
+                {"thread": "1", "relpath": "third"},
+                {"thread": "1", "relpath": "fourth"}
+            ]}""")
+        script.run()
+        worker = script._get_worker("1")
+        self.assertEqual(["first", "second", "third", "fourth"],
+                         [s[0] for s in worker.stats])
+
+    def test_two_threads(self):
+        script = NoActionScript.parse("""{
+            "parameters": {"base_url": ""},
+            "requests": [
+                {"thread": "1", "relpath": "first"},
+                {"thread": "2", "relpath": "second"},
+                {"thread": "1", "relpath": "third"},
+                {"thread": "2", "relpath": "fourth"}
+            ]}""")
+        script.run()
+        worker = script._get_worker("1")
+        self.assertEqual(["first", "third"],
+                         [s[0] for s in worker.stats])
+        worker = script._get_worker("2")
+        self.assertEqual(["second", "fourth"],
+                         [s[0] for s in worker.stats])

